/*
 * YAFFS: Yet another FFS. A NAND-flash specific file system.
 * ynandif.c: NAND interface functions for the WinCE port of YAFFS.
 *
 * Copyright (C) 2002 Trimble Navigation Ltd.
 *
 * Created by Brad Beveridge <brad.beveridge@trimble.co.nz>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 * General Public License for more details. You should have received a 
 * copy of the GNU General Public License along with this program; 
 * if not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 *
 * $Id: ynandif.c,v 1.1 2002-11-08 07:30:00 charles Exp $
 */
#include "ynandif.h"
#include <windows.h>

HANDLE devHandle = 0;

/*
*	Functions that need to be provided for YAFFS
*/
int ynandif_WriteChunkToNAND(yaffs_Device *dev, int chunkInNAND,const __u8 *data, yaffs_Spare *spare)
{
	if (devHandle)
	{
		ynandif_data writeData;
		int result;

		writeData.chunk = chunkInNAND;
		writeData.data = (__u8 *)data;
		writeData.spare = (char *)spare;
		
		if (!DeviceIoControl(devHandle,
			 			     YNANDIF_WRITE,
						     &writeData,
						     sizeof(ynandif_data),
						     &result,
						     sizeof(result),
						     NULL,
							 NULL))
			return YAFFS_FAIL;

		return result;
	}

	return YAFFS_FAIL;
}

int ynandif_ReadChunkFromNAND(yaffs_Device *dev,int chunkInNAND, __u8 *data, yaffs_Spare *spare)
{
	if (devHandle)
	{
		ynandif_data readData;
		int result;

		readData.chunk = chunkInNAND;
		readData.data = data;
		readData.spare = (char *)spare;
		
		if (!DeviceIoControl(devHandle,
			 			     YNANDIF_READ,
						     &readData,
						     sizeof(ynandif_data),
						     &result,
						     sizeof(result),
						     NULL,
							 NULL))
			return YAFFS_FAIL;

		return result;
	}

	return YAFFS_FAIL;
}

int ynandif_EraseBlockInNAND(yaffs_Device *dev, int blockNumber)
{
	if (devHandle)
	{
		int result;
		
		if (!DeviceIoControl(devHandle,
			 			     YNANDIF_ERASE,
						     &blockNumber,
						     sizeof(int),
						     &result,
						     sizeof(result),
						     NULL,
							 NULL))
			return YAFFS_FAIL;

		return result;
	}

// if we return YAFFS_FAIL, then yaffs will retire this block & mark it bad, not exactly
// what we want to do by default.
	return YAFFS_OK;
}

int ynandif_InitialiseNAND(yaffs_Device *dev)
{
	RETAILMSG(1, (L"ynandif_InitialiseNAND\r\n"));
	devHandle = CreateFile(L"YND1:",
						   GENERIC_READ|GENERIC_WRITE,
						   0,
						   NULL,
						   OPEN_EXISTING,
						   0,
						   0);

	if (!devHandle)
		return 0;

	RETAILMSG(1, (L"devhandle open\r\n"));

	RETAILMSG(1, (L"DeviceIo INIT\r\n"));
	if (!DeviceIoControl(devHandle,
 						 YNANDIF_INIT,
						 NULL,
						 0,
						 NULL,
						 0,
						 NULL,
						 NULL))
		return 0;


	if (dev)
	{
		int nBlocks = ynandif_GetChipSize(0xFF) / (YAFFS_CHUNKS_PER_BLOCK * YAFFS_BYTES_PER_CHUNK);
		dev->startBlock = 1;  // Don't use block 0
		dev->endBlock = nBlocks - 1;
	}

	return 1;
}

int ynandif_EraseAllBlocks(yaffs_Device *dev)
{
	int numBlocks, counter;
	numBlocks = ynandif_GetChipSize(0xFF) / (YAFFS_CHUNKS_PER_BLOCK * YAFFS_BYTES_PER_CHUNK);
	for (counter = 0; counter < numBlocks; counter++)
	{
		ynandif_EraseBlockInNAND(dev, counter);
	}
	return YAFFS_OK;
}

int ynandif_GetChipSize(unsigned char chipNumber)
{
	int ret = 0;
	RETAILMSG(1, (L"DeviceIo GETSIZE\r\n"));
	if (devHandle)
	{
		RETAILMSG(1, (L"DeviceIo GETSIZE - getting ret\r\n"));
		DeviceIoControl(devHandle,
					YNANDIF_GETSIZE,
					NULL,
					0,
					&ret,
					sizeof(ret),
					NULL,
					NULL);
	}
	RETAILMSG(1, (L"DeviceIo GETSIZE ret %X\r\n", ret));



	return ret;
}
